@@ -1,7 +1,6 @@
 import asyncio
-from unittest import mock
 
-import asynctest
+import mock
 import pytest
 from aioredis import (
     ConnectionClosedError,
@@ -26,22 +25,32 @@ from aioredis_cluster.errors import (
 from aioredis_cluster.structs import Address
 
 
+def create_async_mock(**kwargs):
+    if "return_value" not in kwargs:
+        kwargs["return_value"] = mock.Mock()
+    return mock.AsyncMock(**kwargs)
+
+
 def get_pooler_mock():
     mocked = mock.NonCallableMock()
-    mocked.ensure_pool = asynctest.CoroutineMock()
-    mocked.close_only = asynctest.CoroutineMock()
+    mocked.ensure_pool = create_async_mock()
+    mocked.close_only = create_async_mock()
 
     conn = mock.NonCallableMock()
-    conn.execute = asynctest.CoroutineMock()
+    conn.execute = create_async_mock()
 
     pool = mocked.ensure_pool.return_value
-    pool.execute = asynctest.CoroutineMock()
-    pool.auth = asynctest.CoroutineMock()
-    pool.acquire = asynctest.CoroutineMock(return_value=conn)
+    pool.execute = create_async_mock()
+    pool.auth = create_async_mock()
+    pool.acquire = create_async_mock(return_value=conn)
+    pool.maxsize = 10
+    pool.minsize = 1
+    pool.size = 1
+    pool.freesize = 0
 
     conn_acquirer = mock.MagicMock()
-    conn_acquirer.__aenter__ = asynctest.CoroutineMock(return_value=conn)
-    conn_acquirer.__aexit__ = asynctest.CoroutineMock(return_value=None)
+    conn_acquirer.__aenter__ = create_async_mock(return_value=conn)
+    conn_acquirer.__aexit__ = create_async_mock(return_value=None)
 
     pool.get.return_value = conn_acquirer
 
@@ -53,7 +62,7 @@ def get_pooler_mock():
 
 def get_manager_mock():
     mocked = mock.NonCallableMock()
-    mocked.get_state = asynctest.CoroutineMock()
+    mocked.get_state = create_async_mock(return_value=mock.NonCallableMock())
     mocked.commands = default_registry
 
     return mocked
@@ -83,7 +92,7 @@ async def test_init__defaults(mocker):
     assert cl._pooler is mocked_pooler.return_value
     assert cl._manager is mocked_manager.return_value
     assert cl._attempt_timeout == cl.ATTEMPT_TIMEOUT
-    mocked_pooler.assert_called_once_with(cl._create_pool, reap_frequency=None)
+    mocked_pooler.assert_called_once_with(cl._create_default_pool, reap_frequency=None)
     mocked_manager.assert_called_once_with(
         ["addr1", "addr2"],
         cl._pooler,
@@ -129,7 +138,7 @@ async def test_init__customized(mocker):
     assert cl._manager is mocked_manager.return_value
     assert cl._attempt_timeout == kwargs["attempt_timeout"]
     mocked_pooler.assert_called_once_with(
-        cl._create_pool, reap_frequency=kwargs["idle_connection_timeout"]
+        cl._create_default_pool, reap_frequency=kwargs["idle_connection_timeout"]
     )
     mocked_manager.assert_called_once_with(
         ["addr1", "addr2"],
@@ -140,9 +149,7 @@ async def test_init__customized(mocker):
 
 
 async def test_create_pool(mocker):
-    mocked_create_pool = mocker.patch(
-        Cluster.__module__ + ".create_pool", new=asynctest.CoroutineMock()
-    )
+    mocked_create_pool = mocker.patch(Cluster.__module__ + ".create_pool", new=create_async_mock())
 
     pool_cls = mock.Mock()
     cl = Cluster(
@@ -170,6 +177,64 @@ async def test_create_pool(mocker):
     )
 
 
+async def test_create_pool_by_addr(mocker):
+    mocked_create_pool = mocker.patch(Cluster.__module__ + ".create_pool", new=create_async_mock())
+
+    pool_cls = mock.Mock()
+    cl = Cluster(
+        ["addr1"],
+        password="foobar",
+        encoding="utf-8",
+        pool_minsize=5,
+        pool_maxsize=15,
+        connect_timeout=1.2,
+        pool_cls=pool_cls,
+    )
+    mocked_manager = mocker.patch.object(cl, "_manager", new=get_manager_mock())
+    state = mocked_manager.get_state.return_value
+    state._data.masters = [
+        mock.NonCallableMock(name="master1", addr=("addr1", 8001)),
+        mock.NonCallableMock(name="master2", addr=("addr2", 8002)),
+        mock.NonCallableMock(name="master3", addr=("addr3", 8003)),
+    ]
+
+    addr = Address("addr2", 8002)
+    redis = await cl.create_pool_by_addr(addr, maxsize=42)
+
+    assert redis.connection is mocked_create_pool.return_value
+    mocked_create_pool.assert_called_once_with(
+        ("addr2", 8002),
+        pool_cls=pool_cls,
+        password="foobar",
+        encoding="utf-8",
+        minsize=5,
+        maxsize=42,
+        create_connection_timeout=1.2,
+    )
+
+
+async def test_create_pool_by_addr__bad_addr(mocker):
+    pool_cls = mock.Mock()
+    cl = Cluster(
+        ["addr1"],
+        password="foobar",
+        encoding="utf-8",
+        pool_minsize=5,
+        pool_maxsize=15,
+        connect_timeout=1.2,
+        pool_cls=pool_cls,
+    )
+    mocked_manager = mocker.patch.object(cl, "_manager", new=get_manager_mock())
+    state = mocked_manager.get_state.return_value
+    state.has_addr.return_value = False
+
+    addr = Address("addr4", 8004)
+    with pytest.raises(ValueError):
+        await cl.create_pool_by_addr(addr, maxsize=42)
+
+    state.has_addr.assert_called_once_with(addr)
+
+
 async def test_auth(mocker):
     cl = Cluster(["addr1"])
 
@@ -180,7 +245,7 @@ async def test_auth(mocker):
         for p in [pool_mock, pool_mock]:
             await fn(p)
 
-    mocked_pooler.batch_op = asynctest.CoroutineMock(side_effect=batch_op_se)
+    mocked_pooler.batch_op = create_async_mock(side_effect=batch_op_se)
 
     await cl.auth("PASSWORD")
 
@@ -274,7 +339,7 @@ async def test_execute__success_one_connection_retry(mocker):
     state = mocked_manager.get_state.return_value
 
     mocked_execute_retry_slowdown = mocker.patch.object(
-        cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock()
+        cl, "_execute_retry_slowdown", new=create_async_mock()
     )
 
     result = await cl.execute("get", "key")
@@ -341,7 +406,7 @@ async def test_execute__success_several_problem_retry(mocker):
     state = mocked_manager.get_state.return_value
 
     mocked_execute_retry_slowdown = mocker.patch.object(
-        cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock()
+        cl, "_execute_retry_slowdown", new=create_async_mock()
     )
 
     result = await cl.execute("get", "key")
@@ -393,7 +458,7 @@ async def test_execute__cancelled_error(mocker):
     mocked_manager = mocker.patch.object(cl, "_manager", new=get_manager_mock())
 
     mocked_execute_retry_slowdown = mocker.patch.object(
-        cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock()
+        cl, "_execute_retry_slowdown", new=create_async_mock()
     )
 
     with pytest.raises(asyncio.CancelledError):
@@ -440,7 +505,7 @@ async def test_execute__error_after_max_attempts(mocker):
     state = mocked_manager.get_state.return_value
 
     mocked_execute_retry_slowdown = mocker.patch.object(
-        cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock()
+        cl, "_execute_retry_slowdown", new=create_async_mock()
     )
 
     with pytest.raises(MovedError) as ei:
@@ -476,7 +541,7 @@ async def test_execute__unexpected_error(mocker):
     mocked_manager = mocker.patch.object(cl, "_manager", new=get_manager_mock())
 
     mocked_execute_retry_slowdown = mocker.patch.object(
-        cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock()
+        cl, "_execute_retry_slowdown", new=create_async_mock()
     )
 
     with pytest.raises(RuntimeError, match="Boom!"):
@@ -497,7 +562,7 @@ async def test_execute__retriable_error_in_unexpected_call(mocker):
     mocked_manager.get_state.side_effect = ConnectionRefusedError()
 
     mocked_execute_retry_slowdown = mocker.patch.object(
-        cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock()
+        cl, "_execute_retry_slowdown", new=create_async_mock()
     )
 
     with pytest.raises(ConnectionRefusedError):
@@ -558,7 +623,7 @@ async def test_execute__retryable_errors(mocker, error, retry_node_addr, require
     state.random_slot_replica.return_value.addr = Address("random_replica", 6379)
 
     mocked_execute_retry_slowdown = mocker.patch.object(
-        cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock()
+        cl, "_execute_retry_slowdown", new=create_async_mock()
     )
 
     result = await cl.execute("get", "key")
@@ -618,7 +683,7 @@ async def test_keys_master__with_retries_but_success(mocker):
     mocked_manager = mocker.patch.object(cl, "_manager", new=get_manager_mock())
     state = mocked_manager.get_state.return_value
 
-    mocker.patch.object(cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock())
+    mocker.patch.object(cl, "_execute_retry_slowdown", new=create_async_mock())
 
     result = await cl.keys_master("key")
 
@@ -645,7 +710,7 @@ async def test_keys_master__with_error_after_retries(mocker):
     mocked_manager = mocker.patch.object(cl, "_manager", new=get_manager_mock())
     state = mocked_manager.get_state.return_value
 
-    mocker.patch.object(cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock())
+    mocker.patch.object(cl, "_execute_retry_slowdown", new=create_async_mock())
 
     with pytest.raises(ClusterDownError):
         await cl.keys_master("key")
@@ -680,31 +745,31 @@ async def test_all_masters__with_error_after_retries(mocker):
 
     mocked_manager = mocker.patch.object(cl, "_manager", new=get_manager_mock())
     state = mocked_manager.get_state.return_value
-    state.masters = [
+    state._data.masters = [
         mock.NonCallableMock(name="master1"),
         mock.NonCallableMock(name="master2"),
         mock.NonCallableMock(name="master3"),
     ]
 
-    mocker.patch.object(cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock())
+    mocker.patch.object(cl, "_execute_retry_slowdown", new=create_async_mock())
 
     with pytest.raises(ClusterDownError):
         await cl.all_masters()
 
 
-async def test_execute__with_attempt_timeout__non_idempotent(mocker, loop):
+async def test_execute__with_attempt_timeout__non_idempotent(mocker, event_loop):
     cl = Cluster(["addr1"], attempt_timeout=0.001)
 
     mocked_pooler = mocker.patch.object(cl, "_pooler", new=get_pooler_mock())
     pool = mocked_pooler._pool
-    pool.execute = mock.Mock(return_value=loop.create_future())
+    pool.execute = mock.Mock(return_value=event_loop.create_future())
 
     mocked_manager = mocker.patch.object(cl, "_manager", new=get_manager_mock())
     state = mocked_manager.get_state.return_value
     state.slot_master.return_value.addr = Address("1.2.3.4", 9999)
 
     mocked_execute_retry_slowdown = mocker.patch.object(
-        cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock()
+        cl, "_execute_retry_slowdown", new=create_async_mock()
     )
 
     with pytest.raises(asyncio.TimeoutError):
@@ -716,7 +781,7 @@ async def test_execute__with_attempt_timeout__non_idempotent(mocker, loop):
     mocked_execute_retry_slowdown.assert_not_called()
 
 
-async def test_execute__with_attempt_timeout__idempotent(mocker, loop):
+async def test_execute__with_attempt_timeout__idempotent(mocker):
     cl = Cluster(["addr1"], max_attempts=3)
 
     mocked_pooler = mocker.patch.object(cl, "_pooler", new=get_pooler_mock())
@@ -729,7 +794,7 @@ async def test_execute__with_attempt_timeout__idempotent(mocker, loop):
     state.random_node.return_value.addr = Address("6.6.6.6", 9999)
 
     mocked_execute_retry_slowdown = mocker.patch.object(
-        cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock()
+        cl, "_execute_retry_slowdown", new=create_async_mock()
     )
 
     with pytest.raises(asyncio.TimeoutError):
@@ -741,7 +806,7 @@ async def test_execute__with_attempt_timeout__idempotent(mocker, loop):
     assert mocked_execute_retry_slowdown.call_count == 2
 
 
-async def test_execute__with_attempt_timeout__idempotent__success(mocker, loop):
+async def test_execute__with_attempt_timeout__idempotent__success(mocker):
     cl = Cluster(["addr1"], max_attempts=4)
 
     mocked_pooler = mocker.patch.object(cl, "_pooler", new=get_pooler_mock())
@@ -759,7 +824,7 @@ async def test_execute__with_attempt_timeout__idempotent__success(mocker, loop):
     state.random_node.return_value.addr = Address("6.6.6.6", 9999)
 
     mocked_execute_retry_slowdown = mocker.patch.object(
-        cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock()
+        cl, "_execute_retry_slowdown", new=create_async_mock()
     )
 
     result = await cl.execute("get", "key")
@@ -771,7 +836,7 @@ async def test_execute__with_attempt_timeout__idempotent__success(mocker, loop):
     assert mocked_execute_retry_slowdown.call_count == 3
 
 
-async def test_keys_master__with_attempt_timeout(mocker, loop):
+async def test_keys_master__with_attempt_timeout(mocker):
     cl = Cluster(["addr1"], max_attempts=2)
 
     mocked_pooler = mocker.patch.object(cl, "_pooler", new=get_pooler_mock())
@@ -783,7 +848,7 @@ async def test_keys_master__with_attempt_timeout(mocker, loop):
     state.slot_master.return_value.addr = Address("1.2.3.4", 9999)
 
     mocked_execute_retry_slowdown = mocker.patch.object(
-        cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock()
+        cl, "_execute_retry_slowdown", new=create_async_mock()
     )
 
     with pytest.raises(asyncio.TimeoutError):
@@ -795,25 +860,31 @@ async def test_keys_master__with_attempt_timeout(mocker, loop):
     assert mocked_execute_retry_slowdown.call_count == 1
 
 
-async def test_all_masters__with_attempt_timeout(mocker, loop):
+async def test_all_masters__with_attempt_timeout(mocker, event_loop):
     cl = Cluster(["addr1"], attempt_timeout=0.001)
 
     mocked_pooler = mocker.patch.object(cl, "_pooler", new=get_pooler_mock())
     pool = mocked_pooler._pool
-    execute1_fut = loop.create_future()
+    execute1_fut = event_loop.create_future()
     execute1_fut.set_result(None)
-    execute2_fut = loop.create_future()
-    pool.execute.side_effect = [execute1_fut, execute2_fut]
+    execute2_fut = event_loop.create_future()
+
+    execute_futs_iter = iter([execute1_fut, execute2_fut])
+
+    async def execute_side_effect(*args, **kwargs):
+        return await next(execute_futs_iter)
+
+    pool.execute.side_effect = execute_side_effect
 
     mocked_manager = mocker.patch.object(cl, "_manager", new=get_manager_mock())
     state = mocked_manager.get_state.return_value
-    state.masters = [
+    state._data.masters = [
         mock.NonCallableMock(addr=Address("1.2.3.4", 6666)),
         mock.NonCallableMock(addr=Address("1.2.3.4", 9999)),
     ]
 
     mocked_execute_retry_slowdown = mocker.patch.object(
-        cl, "_execute_retry_slowdown", new=asynctest.CoroutineMock()
+        cl, "_execute_retry_slowdown", new=create_async_mock()
     )
 
     with pytest.raises(asyncio.TimeoutError):
